/*
 *  Copyright (C) 2010 Luca Deri <deri@ntop.org>
 *
 *  		       http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "nprobe.h"

#define BASE_ID            NTOP_BASE_ID+205
#define QUERY_MAX_LEN      32

#define MAX_NUM_FILES      10000 
#define MAX_NUM_LINES      10000 


/* Ports */
#define DNS_PORT             53


static V9V10TemplateElementId dnsPlugin_template[] = {
  { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID,   VARIABLE_FIELD_LEN, QUERY_MAX_LEN, ascii_format, dump_as_ascii, "DNS_QUERY", "DNS QUERY" },
  { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID+1, STATIC_FIELD_LEN, 2, numeric_format, dump_as_uint, "DNS_QUERY_ID", "DNS query transaction Id" },
  { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID+2, STATIC_FIELD_LEN, 1, numeric_format, dump_as_uint, "DNS_QUERY_TYPE", "DNS query type (e.g. 1=A, 2=NS..)" },
  { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID+3, STATIC_FIELD_LEN, 1, numeric_format, dump_as_uint, "DNS_RET_CODE", "DNS return code (e.g. 0=no error)" },
  { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID+4, STATIC_FIELD_LEN, 1, numeric_format, dump_as_uint, "DNS_NUM_ANSWER", "DNS # of returned answers" },
  { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, 0, STATIC_FIELD_LEN, 0, 0, 0, NULL, NULL }
};

struct dns_packet_header {
  u_int16_t transaction_id, flags, num_queries, answer_rrs, authority_rrs, additional_rrs;
};

struct plugin_info {
  char *host_query;
  u_int8_t num_query_pkts, num_answer_pkts;
  u_int8_t ret_code, num_answer, query_type;
  u_int16_t transaction_id;
};

//#define DEBUG

static char last_dns_dump_dir[256] = { 0 };
static char dns_exec_command[256] = { 0 };
static u_int dns_file_dump_timeout  = 300;  /* 5 min */
static u_int dns_directory_duration = 3600; /* 1 hour */

/* *********************************************** */

static PluginInfo dnsPlugin; /* Forward */
static FILE *fd = NULL;
static time_t fd_close_timeout = 0;
static u_int num_file = 0, num_log_lines = 0;
static char dns_dump_dir[256] = { 0 };

/* ******************************************* */

void dnsPlugin_init(int argc, char *argv[]) {
  int i;

  traceEvent(TRACE_INFO, "Initialized DNS plugin");

  for(i=0; i<argc; i++)
    if(!strcmp(argv[i], "--dns-dump-dir")) {
      if((i+1) < argc) {
	int len;
	
	snprintf(dns_dump_dir, sizeof(dns_dump_dir), "%s", argv[i+1]);
	len = strlen(dns_dump_dir);
	if(len > 0) len--;
	
	if(dns_dump_dir[len] == '/')
	  dns_dump_dir[len] = '\0';

	traceEvent(TRACE_NORMAL, "DNS log files will be saved in %s", dns_dump_dir);
      }
    } else if(!strcmp(argv[i], "--dns-exec-cmd")) {
      if((i+1) < argc) {
	snprintf(dns_exec_command, sizeof(dns_exec_command), "%s", argv[i+1]);
	traceEvent(TRACE_NORMAL, "DNS directories will be processed by '%s'", 
		   dns_exec_command);
      }
    }
}

/* *********************************************** */

/* Handler called whenever an incoming packet is received */

static void dnsPlugin_packet(u_char new_bucket, void *pluginData,
			      FlowHashBucket* bkt,
			     int flow_direction,
			      u_short proto, u_char isFragment,
			      u_short numPkts, u_char tos,
			      u_short vlanId, struct eth_header *ehdr,
			      IpAddress *src, u_short sport,
			      IpAddress *dst, u_short dport,
			      u_int len, u_int8_t flags, u_int8_t icmpType,
			      u_short numMplsLabels,
			      u_char mplsLabels[MAX_NUM_MPLS_LABELS][MPLS_LABEL_LEN],
			      const struct pcap_pkthdr *h, const u_char *p,
			      u_char *payload, int payloadLen) {
  PluginInformation *info;
  struct plugin_info *pinfo;
  u_int8_t direction, is_query = 0, ret_code = 0;

  if(new_bucket) {
    info = (PluginInformation*)malloc(sizeof(PluginInformation));
    if(info == NULL) {
      traceEvent(TRACE_ERROR, "Not enough memory?");
      return; /* Not enough memory */
    }

    info->pluginPtr  = (void*)&dnsPlugin;
    pluginData = info->pluginData = malloc(sizeof(struct plugin_info));

    if(info->pluginData == NULL) {
      traceEvent(TRACE_ERROR, "Not enough memory?");
      free(info);
      return; /* Not enough memory */
    } else
      memset(info->pluginData, 0, sizeof(struct plugin_info));

    info->next = bkt->plugin;
    bkt->plugin = info;
  }

  /*
    Do not put the checks below at the beginning of the function as otherwise
    the flow will not be able to export info about the template specified
   */
  if(proto != IPPROTO_UDP) return;
 
  if(sport == DNS_PORT)
    direction = 0; /* server -> client */
  else if(dport == DNS_PORT)
    direction = 1; /* client -> server */
  else
    return;
  
  pinfo = (struct plugin_info*)pluginData;
  
  if(payloadLen > (sizeof(struct dns_packet_header)+4)) {
    struct dns_packet_header *header = (struct dns_packet_header*)payload;
    char hostname[256]= { 0 };
    int i = sizeof(struct dns_packet_header)+1, j = 0;

    //traceEvent(TRACE_INFO, "==> [len=%d][%s]", payloadLen, payload);
    header->transaction_id = ntohs(header->transaction_id);
    header->num_queries = ntohs(header->num_queries);
    header->answer_rrs = ntohs(header->answer_rrs);
    header->flags = ntohs(header->flags);
    is_query = (header->flags & 0x8000) ? 0 : 1;
    ret_code = is_query ? 0 : (header->flags & 0x0F);

    while((i < payloadLen) && (payload[i] != '\0')) {
      hostname[j] = payload[i];
      if(hostname[j] < ' ') hostname[j] = '.';
      j++, i++;
    }

    traceEvent(TRACE_INFO, "[isQuery=%d][flags=0x%X][transaction_id=0x%X][num_queries=%u][answer_rrs=%u][ret_code=%d][%s]",
	       is_query, header->flags, header->transaction_id, header->num_queries, header->answer_rrs, ret_code, hostname);

    if(is_query) {
      u_int16_t query_type = 0;

      i += 2; /* Skip query class */
      memcpy(&query_type, &payload[i], 2); 
      i += 2, query_type  = ntohs(query_type);
      
      pinfo->num_query_pkts++, pinfo->query_type = query_type;
      if(pinfo->host_query) { free(pinfo->host_query); pinfo->host_query = NULL; }
    } else {

    }
  }
}

/* *********************************************** */

static void dnsPlugin_dump_flow(FlowHashBucket* bkt, void *pluginData) {
#if 0
  if(pluginData != NULL) {
    struct plugin_info *pinfo = (struct plugin_info*)pluginData;

    if(pinfo->dns_url) {
#if 0
      char buf[64], when[64], *host;
      struct tm tm;
      time_t when_time;

      host = _intoa((bkt->dport == 80) ? bkt->src->host : bkt->dst->host, buf, sizeof(buf));
      when_time = (bkt->dport == 80) ? bkt->flowTimers.firstSeenSent.tv_sec : bkt->flowTimers.firstSeenRcvd.tv_sec;
      /* 80.8.55.1 - - [01/Jan/2001:00:00:10 +0100] "GET /page1.html DNS/1.0" 200 7009 "-" "Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0)" */
      localtime_r(&when_time, &tm);

      strftime(when, sizeof(when), "%d/%B/%Y:%H:%M:%S %z", &tm);

      printf("%s - - [%s] \"%s %s\" %d %u \"-\" \"%s\" \"%s\"\n",
	     host, when, method2str(pinfo->method), pinfo->dns_url, pinfo->ret_code,
	     (bkt->flowCounters.bytesSent + bkt->flowCounters.bytesRcvd),
	     pinfo->referer ? pinfo->referer : "",
	     pinfo->user_agent ? pinfo->user_agent : "");
#else
      if(dns_dump_dir[0] != '\0') {
	char buf[64];
	char *client = _intoa((bkt->dport == 80) ? bkt->src->host : bkt->dst->host, buf, sizeof(buf));
	u_int begin_time = (bkt->dport == 80) ? bkt->flowTimers.firstSeenSent.tv_sec : bkt->flowTimers.firstSeenRcvd.tv_sec;
	u_int end_time = (bkt->dport == 80) ? bkt->flowTimers.lastSeenSent.tv_sec : bkt->flowTimers.lastSeenRcvd.tv_sec;
	time_t now = time(NULL);
	if(pinfo->dns_host == NULL) {
	  char buf1[64];

	  pinfo->dns_url = strdup(_intoa((bkt->sport == 80) ? bkt->src->host : bkt->dst->host, buf1, sizeof(buf1))); 
	}

	/* 192.168.0.200|www.macintouch.com|GET|/images/filewave02.gif|200|www.macintouch.com/|(null)||25505|1133966827|1133966829 */

	if(fd && (fd_close_timeout < now)) {
	  fclose(fd);
	  fd = NULL;
	}

	if(fd == NULL) {
	  char current_dump_dir[256], creation_time[256];
	  time_t theTime = time(NULL);
	  struct tm *tm;
	  int rc = 0;
	  
	  theTime -= (theTime % dns_directory_duration);
	  tm = localtime(&theTime);
	  
	  strftime(creation_time, sizeof(creation_time), "%Y/%m/%d/%H", tm);
	  snprintf(current_dump_dir, sizeof(current_dump_dir), "%s%c%s",
		   dns_dump_dir, CONST_DIR_SEP, creation_time);
	  
	  if(strcmp(last_dns_dump_dir, current_dump_dir)) {
	    exec_dns_cmd();
	    rc = mkdir_p(current_dump_dir);
	    num_file = 0;
	    snprintf(last_dns_dump_dir, sizeof(last_dns_dump_dir), "%s", current_dump_dir);
	  }
      
	  if(rc == 0) {
	    snprintf(buf, sizeof(buf), "%s/dns_%d.txt",
		     current_dump_dir, num_file++);
	    num_log_lines = 0;
	    fd = fopen(buf, "w");

	    if(fd) {
	      traceEvent(TRACE_INFO, "Created %s", buf);
	      fprintf(fd, "#\n# %s\t%s\t%s\t%s\t%s\t"
		      "%s\t%s\t%s\t"
		      "%s\t%s\t%s\t%s\n#\n",
		      "Client", "Server", "Method", "URL", "DNSReturnCode",
		      "Referer", "UserAgent", "ContentType",
		      "Bytes", "BeginTime", "EndTime", "Flow Hash");
	      fd_close_timeout = now + dns_file_dump_timeout;
	    } else
	      traceEvent(TRACE_WARNING, "Unable to create file %s [%s]", buf, strerror(errno));
	  }
	}

	if(fd) {
	  u_int32_t flow_hash, srcHost, dstHost;
	  
	  if(bkt->src->host.ipVersion == 4) {
	    srcHost = bkt->src->host.ipType.ipv4, dstHost = bkt->dst->host.ipType.ipv4;
	  } else {
	    srcHost = bkt->src->host.ipType.ipv6.s6_addr32[0]+bkt->src->host.ipType.ipv6.s6_addr32[1]
	      +bkt->src->host.ipType.ipv6.s6_addr32[2]+bkt->src->host. ipType.ipv6.s6_addr32[3];
	    dstHost = bkt->dst->host.ipType.ipv6.s6_addr32[0]+bkt->dst->host.ipType.ipv6.s6_addr32[1]
	      +bkt->dst->host.ipType.ipv6.s6_addr32[2]+bkt->dst->host.ipType.ipv6.s6_addr32[3];
	  }
	  
	  flow_hash = (bkt->proto+srcHost+dstHost+bkt->sport+bkt->dport);

	  fprintf(fd, "%s\t%s\t%s\t%s\t%u\t"
		  "%s\t%s\t%s\t"
		  "%u\t%u\t%u\t%u\n",
		  client, pinfo->dns_host, method2str(pinfo->method), 
		  pinfo->dns_url, pinfo->ret_code,
		  pinfo->referer ? pinfo->referer : "", pinfo->user_agent ? pinfo->user_agent : "", 
		  pinfo->content_type ? pinfo->content_type : "",
		  (bkt->flowCounters.bytesSent + bkt->flowCounters.bytesRcvd), 
		  begin_time, end_time, flow_hash);
	  num_log_lines++;
	  
	  if(num_log_lines == MAX_NUM_LINES) {
	    fclose(fd);
	    fd = NULL;
	  }
	}
      }
#endif
    }
  }
#endif
}

/* *********************************************** */

/* Handler called when the flow is deleted (after export) */

static void dnsPlugin_delete(FlowHashBucket* bkt, void *pluginData) {

  if(pluginData) {
    struct plugin_info *pinfo = (struct plugin_info*)pluginData;

    dnsPlugin_dump_flow(bkt, pluginData);

    if(pinfo->host_query)   free(pinfo->host_query);
    free(pluginData);
  }
}

/* *********************************************** */

/* Handler called at startup when the template is read */

static V9V10TemplateElementId* dnsPlugin_get_template(char* template_name) {
  int i;

  for(i=0; dnsPlugin_template[i].templateElementId != 0; i++) {
    if(!strcmp(template_name, dnsPlugin_template[i].templateElementName)) {
      return(&dnsPlugin_template[i]);
    }
  }

  return(NULL); /* Unknown */
}

/* *********************************************** */

/* Handler called whenever a flow attribute needs to be exported */

static int dnsPlugin_export(void *pluginData, V9V10TemplateElementId *theTemplate,
			     int direction /* 0 = src->dst, 1 = dst->src */,
			     FlowHashBucket *bkt, char *outBuffer,
			     uint* outBufferBegin, uint* outBufferMax) {
  int i;

  if(direction == 0) /* Avoid exporting twice */
    dnsPlugin_dump_flow(bkt, pluginData);

  if(theTemplate == NULL) return(-1);

  for(i=0; dnsPlugin_template[i].templateElementId != 0; i++) {
    if(theTemplate->templateElementId == dnsPlugin_template[i].templateElementId) {
      if((*outBufferBegin)+dnsPlugin_template[i].templateElementLen > (*outBufferMax))
	return(-2); /* Too long */

      if(pluginData) {
	struct plugin_info *info = (struct plugin_info *)pluginData;
	uint len;

	if(info->host_query == NULL) info->host_query = strdup("");

	switch(dnsPlugin_template[i].templateElementId) {
	case BASE_ID:
	  if((readOnlyGlobals.netFlowVersion == 10)
	     && (dnsPlugin_template[i].variableFieldLength == VARIABLE_FIELD_LEN)) {
	    len = min(strlen(info->host_query), dnsPlugin_template[i].templateElementLen);

	    if(len < 255)
	      copyInt8(len, outBuffer, outBufferBegin, outBufferMax);
	    else {
	      copyInt8(255, outBuffer, outBufferBegin, outBufferMax);
	      copyInt16(len, outBuffer, outBufferBegin, outBufferMax);
	    }
	  } else
	    len = dnsPlugin_template[i].templateElementLen;

	  memcpy(&outBuffer[*outBufferBegin], info->host_query, len);
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> HOST='%s'", info->host_query);
#endif
	  (*outBufferBegin) += len;
	  break;

	case BASE_ID+1:
	  copyInt16(info->transaction_id, outBuffer, outBufferBegin, outBufferMax);
	  break;

	case BASE_ID+2:
	  copyInt8(info->query_type, outBuffer, outBufferBegin, outBufferMax);
	  break;

	case BASE_ID+3:
	  copyInt8(info->ret_code, outBuffer, outBufferBegin, outBufferMax);
	  break;

	case BASE_ID+4:
	  copyInt8(info->num_answer, outBuffer, outBufferBegin, outBufferMax);
	  break;

	default:
	  return(-1); /* Not handled */
	}

	return(0);
      }
    }
  }

  return(-1); /* Not handled */
}

/* *********************************************** */

/* Handler called whenever a flow attribute needs to be printed on file */

static int dnsPlugin_print(void *pluginData, V9V10TemplateElementId *theTemplate,
			    int direction /* 0 = src->dst, 1 = dst->src */,
			    FlowHashBucket *bkt, char *line_buffer, uint line_buffer_len) {
  int i;

  printf("==>> dnsPlugin_print()\n");

  for(i=0; dnsPlugin_template[i].templateElementId != 0; i++) {
    if(theTemplate->templateElementId == dnsPlugin_template[i].templateElementId) {

      if(pluginData) {
	struct plugin_info *info = (struct plugin_info *)pluginData;

	switch(dnsPlugin_template[i].templateElementId) {
	case BASE_ID:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s", info->host_query ? info->host_query : "");
	  break;

	case BASE_ID+1:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%u", info->transaction_id);
	  break;

	case BASE_ID+2:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%d", info->query_type);
	  break;

	case BASE_ID+3:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%d", info->ret_code);
	  break;

	case BASE_ID+4:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%d", info->num_answer);
	  break;

	default:
	  return(-1); /* Not handled */
	}

	return(0);
      }
    }
  }

  return(-1); /* Not handled */
}

/* *********************************************** */

static V9V10TemplateElementId* dnsPlugin_conf(void) {
  return(dnsPlugin_template);
}

/* *********************************************** */

static void dnsPlugin_help(void) {
  printf("  --dns-dump-dir <dump dir>                          | Directory where DNS logs will be dumped\n");
}

/* *********************************************** */

/* Plugin entrypoint */
static PluginInfo dnsPlugin = {
  NPROBE_REVISION,
  "DNS Protocol Dissector",
  "0.1",
  "Handle DNS protocol",
  "L.Deri <deri@ntop.org>",
  0 /* not always enabled */, 1, /* enabled */
  dnsPlugin_init,
  NULL,
  dnsPlugin_conf,
  dnsPlugin_delete,
  1, /* call packetFlowFctn for each packet */
  dnsPlugin_packet,
  dnsPlugin_get_template,
  dnsPlugin_export,
  dnsPlugin_print,
  NULL,
  dnsPlugin_help
};

/* *********************************************** */

/* Plugin entry fctn */
#ifdef MAKE_STATIC_PLUGINS
PluginInfo* dnsPluginEntryFctn(void)
#else
  PluginInfo* PluginEntryFctn(void)
#endif
{
  return(&dnsPlugin);
}

