/*
 *  Copyright (C) 2010 Emanuele Tomai <tomasi@cli.di.unipi.it>
 *
 *  		       http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "nprobe.h"

#define BASE_ID            NTOP_BASE_ID+195
#define DEBUG

#define MYSQL_MAX_LEN        32     /* Max lenght for string */
#define MYSQL_SERVER_PORT    3306

/* Capabilities */
#define MYSQL_CAPS_CWDB    0x0008    /* if there is DB name in client logging package (only for protocol >= 4.1) */
#define MYSQL_CAPS_41      0x0200    /* if used protocol >= 4.1 */

/* Command (in "Command Packet") */
#define MYSQL_COMM_INIT_DB     2
#define MYSQL_COMM_QUERY       3

#define MYSQL_NUM_COMM     0x1c
static const char *mysql_command[] = /* Translate command (use with p_data->mysql_command) */
  {
    "SLEEP",                /* 0x00 */
    "Quit",
    "Use Database",
    "Query",
    "Show Fields",
    "Create Database",
    "Drop Database",
    "Refresh",
    "Shutdown",
    "Statistics",
    "Process List",
    "Connect",
    "Kill Server Thread",
    "Dump Debuginfo",
    "Ping",
    "Time",
    "Insert Delayed",
    "Change User",
    "Send Binlog",
    "Send Table",
    "Slave Connect",
    "Register Slave",
    "Prepare Statement",
    "Execute Statement",
    "Send BLOB",
    "Close Statement",
    "Reset Statement",
    "Set Option",
    "Fetch Data",           /* 0x1c: MYSQL_NUM_COMM */
  };

static V9V10TemplateElementId mysqlPlugin_template[] = {
    { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID,   VARIABLE_FIELD_LEN, MYSQL_MAX_LEN, ascii_format, dump_as_ascii, "MYSQL_SERVER_VERSION", "MySQL server version" },
    { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID+1, VARIABLE_FIELD_LEN, MYSQL_MAX_LEN, ascii_format, dump_as_ascii, "MYSQL_USERNAME", "MySQL username" },
    { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID+2, VARIABLE_FIELD_LEN, MYSQL_MAX_LEN, ascii_format, dump_as_ascii, "MYSQL_COMMAND", "MySQL command" },
    { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID+3, VARIABLE_FIELD_LEN, MYSQL_MAX_LEN, ascii_format, dump_as_ascii, "MYSQL_DB", "MySQL database in use" },
    { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID+4, VARIABLE_FIELD_LEN, MYSQL_MAX_LEN, ascii_format, dump_as_ascii, "MYSQL_QUERY", "MySQL Query" },
    { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, BASE_ID+5, STATIC_FIELD_LEN, 2, numeric_format, dump_as_uint, "MYSQL_RESPONSE", "MySQL server response" },
    { FLOW_TEMPLATE, LONG_SNAPLEN, NTOP_ENTERPRISE_ID, 0, STATIC_FIELD_LEN, 0, 0, 0, NULL, NULL }
  };

struct packet_header
{
  u_int32_t packet_length; /* only 24 bit */
  u_char packet_number;
};

struct connection_info
{
  char *mysql_server_version;
  char *mysql_username;
  u_int16_t s_caps;
  u_int16_t c_caps;
  u_int16_t c_ext_caps;
  u_char handshake; /* 0 = not terminated, 1 = handshake complete */
};

struct plugin_data
{
  char *mysql_db;
  u_char mysql_command;
  char *mysql_query;
  int8_t mysql_response;
  struct connection_info *c_info;
};

/* Internal function prototypes */
static void mysql_dissect_handshake(FlowHashBucket *, struct packet_header *, struct plugin_data *, u_char *);
static void mysql_dissect_command(FlowHashBucket *, struct packet_header *, struct plugin_data *, u_char *);

/* *********************************************** */

static PluginInfo mysqlPlugin; /* Forward */

/* ******************************************* */

void
mysqlPlugin_init(int argc, char *argv[])
{
  traceEvent(TRACE_INFO, "Initialized MySQL plugin");
}

/* *********************************************** */

/* Handler called whenever an incoming packet is received */

static void
mysqlPlugin_packet(u_char new_bucket, void *pluginData,
		   FlowHashBucket *bkt,
		   int flow_direction,
		   u_short proto, u_char isFragment,
		   u_short numPkts, u_char tos,
		   u_short vlanId, struct eth_header *ehdr,
		   IpAddress *src, u_short sport,
		   IpAddress *dst, u_short dport,
		   u_int len, u_int8_t flags, u_int8_t icmpType,
		   u_short numMplsLabels,
		   u_char mplsLabels[MAX_NUM_MPLS_LABELS][MPLS_LABEL_LEN],
		   const struct pcap_pkthdr *h, const u_char *p,
		   u_char *payload, int payloadLen)
{
  struct plugin_data *p_data;
  struct connection_info *c_info;

  if ( sport != MYSQL_SERVER_PORT && dport != MYSQL_SERVER_PORT )
    return; /* Nothing to be done */

  p_data = (struct plugin_data*)pluginData;

  if(new_bucket)
    {
      PluginInformation *info;

      if( ! (info = (PluginInformation*)malloc(sizeof(PluginInformation))) )
	{
	  traceEvent(TRACE_ERROR, "Not enough memory?");
	  return; /* Not enough memory */
	}

      info->pluginPtr  = (void*)&mysqlPlugin;
      if ( ! (p_data = info->pluginData = malloc(sizeof(struct plugin_data))) )
	{
	  traceEvent(TRACE_ERROR, "Not enough memory?");
	  free(info);
	  return; /* Not enough memory */
	}
      memset(p_data, 0, sizeof(struct plugin_data));
      p_data->mysql_command = MYSQL_NUM_COMM + 1; /* Unknown */
      p_data->mysql_response = -1; /* Error */


      if ( ! (p_data->c_info = malloc(sizeof(struct connection_info))) )
	{
	  traceEvent(TRACE_ERROR, "Not enough memory?");
	  free(p_data);
	  free(info);
	  return; /* Not enough memory */
	}
      memset(p_data->c_info, 0, sizeof (struct connection_info));

      info->next = bkt->plugin;
      bkt->plugin = info;
    }

  c_info = p_data->c_info;

  if(payloadLen > 0)
    {
      struct packet_header pkt_header;

      /* Keep header info */
      memcpy(&pkt_header.packet_length, payload, 3);
      pkt_header.packet_number = payload[3];
      payload += 4; /* Skip header: payload = MySQL packet */

      if ( ! c_info->handshake )
	mysql_dissect_handshake(bkt, &pkt_header, p_data, payload);
      else
	mysql_dissect_command(bkt, &pkt_header, p_data, payload);
    }
}

/* *********************************************** */

/* Handler called when the flow is deleted (after export) */

static void
mysqlPlugin_delete(FlowHashBucket* bkt, void *pluginData)
{
  struct plugin_data *p_data = (struct plugin_data*)pluginData;

  if(p_data != NULL)
    {
      struct connection_info *c_info = p_data->c_info;

      free(p_data->mysql_db);
      free(p_data->mysql_query);
      free(c_info->mysql_server_version);
      free(c_info->mysql_username);
      free(c_info);

      free(p_data);
    }
}

/* *********************************************** */

/* Handler called at startup when the template is read */

static V9V10TemplateElementId*
mysqlPlugin_get_template(char* template_name)
{
  int i;

  for(i=0; mysqlPlugin_template[i].templateElementId != 0; i++)
    if(!strcmp(template_name, mysqlPlugin_template[i].templateElementName))
      return(&mysqlPlugin_template[i]);

  return(NULL); /* Unknown */
}

/* *********************************************** */

/* Handler called whenever a flow attribute needs to be exported */

static int
mysqlPlugin_export(void *pluginData, V9V10TemplateElementId *theTemplate,
		   int direction /* 0 = src->dst, 1 = dst->src */,
		   FlowHashBucket *bkt, char *outBuffer,
		   uint* outBufferBegin, uint* outBufferMax) {
  int i;

  if(theTemplate == NULL) return(-1);

  for(i=0; mysqlPlugin_template[i].templateElementId != 0; i++)
    {
      if(theTemplate->templateElementId == mysqlPlugin_template[i].templateElementId)
	{
	  struct plugin_data *p_data = (struct plugin_data *)pluginData;

	  if((*outBufferBegin)+mysqlPlugin_template[i].templateElementLen > (*outBufferMax))
	    return(-2); /* Too long */

	  if(p_data)
	    {
	      char sql_buf[512] = { 0 };
	      uint len;

	      switch(mysqlPlugin_template[i].templateElementId)
		{
		case BASE_ID+4:
		  snprintf(sql_buf, sizeof(sql_buf), "%s", p_data->mysql_query ? p_data->mysql_query : "");

		  if((readOnlyGlobals.netFlowVersion == 10)
		     && (mysqlPlugin_template[i].variableFieldLength == VARIABLE_FIELD_LEN))
		    {
		      len = min(strlen(sql_buf), mysqlPlugin_template[i].templateElementLen);

		      if(len < 255)
			copyInt8(len, outBuffer, outBufferBegin, outBufferMax);
		      else
			{
			  copyInt8(255, outBuffer, outBufferBegin, outBufferMax);
			  copyInt16(len, outBuffer, outBufferBegin, outBufferMax);
			}
		    }
		  else
		    len = mysqlPlugin_template[i].templateElementLen;

		  memcpy(&outBuffer[*outBufferBegin], sql_buf, len);
#ifdef DEBUG
		  traceEvent(TRACE_INFO, "==> Query='%s'", sql_buf);
#endif
		  (*outBufferBegin) += len;
		  break;
		case BASE_ID+5:
		  copyInt16(p_data->mysql_response, outBuffer, outBufferBegin, outBufferMax);
#ifdef DEBUG
		  traceEvent(TRACE_INFO, "==> Response='%d'", p_data->mysql_response);
#endif
		  break;
		default:
		  return(-1); /* Not handled */
		}

	      return(0);
	    }
	}
    }

  return(-1); /* Not handled */
}

/* *********************************************** */

/* Handler called whenever a flow attribute needs to be printed on file */

static int
mysqlPlugin_print(void *pluginData, V9V10TemplateElementId *theTemplate,
		  int direction /* 0 = src->dst, 1 = dst->src */,
		  FlowHashBucket *bkt, char *line_buffer, uint line_buffer_len)
{
  int i;

  for(i=0; mysqlPlugin_template[i].templateElementId != 0; i++)
    {
      if(theTemplate->templateElementId == mysqlPlugin_template[i].templateElementId)
	{
	  struct plugin_data *p_data = (struct plugin_data *)pluginData;
	  if(p_data)
	    {
	      struct connection_info *c_info = p_data->c_info;

	      switch(mysqlPlugin_template[i].templateElementId)
		{
		case BASE_ID:
		  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s",
			   c_info->mysql_server_version ? c_info->mysql_server_version : "");
		  break;
		case BASE_ID+1:
		  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s",
			   c_info->mysql_username ? c_info->mysql_username : "");
		  break;
		case BASE_ID+2:
		  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s",
			   ( p_data->mysql_command <= MYSQL_NUM_COMM ) ? mysql_command[p_data->mysql_command] : "");
		  break;
		case BASE_ID+3:
		  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s",
			   p_data->mysql_db ? p_data->mysql_db : "");
		  break;
		case BASE_ID+4:
		  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s",
			   p_data->mysql_query ? p_data->mysql_query : "");
		  break;
		case BASE_ID+5:
		  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%d", p_data->mysql_response);
		  break;
		default:
		  return(-1); /* Not handled */
		}

	      return(0);
	    }
	}
    }

  return(-1); /* Not handled */
}

/* *********************************************** */

static V9V10TemplateElementId*
mysqlPlugin_conf(void)
{
  return(mysqlPlugin_template);
}

/* *********************************************** */

/* Plugin entrypoint */
static PluginInfo mysqlPlugin =
  {
    NPROBE_REVISION,
    "MySQL Protocol Dissector",
    "0.1",
    "Handle MySQL protocol",
    "E.Tomasi <tomasi@cli.di.unipi.it>",
    0 /* not always enabled */, 1, /* enabled */
    mysqlPlugin_init,
    NULL, /* Term */
    mysqlPlugin_conf,
    mysqlPlugin_delete,
    1, /* call packetFlowFctn for each packet */
    mysqlPlugin_packet,
    mysqlPlugin_get_template,
    mysqlPlugin_export,
    mysqlPlugin_print,
    NULL,
    NULL
  };

/* *********************************************** */

/* Plugin entry fctn */
#ifdef MAKE_STATIC_PLUGINS
PluginInfo* mysqlPluginEntryFctn(void)
#else
PluginInfo* PluginEntryFctn(void)
#endif
{
  return(&mysqlPlugin);
}

/* *********************************************** */

/* Internal functions */

/* Used by mysqlPlugin_packet() */
static void
mysql_dissect_handshake(FlowHashBucket *bkt, struct packet_header *pkt_header, struct plugin_data *p_data, u_char *packet)
{
  int i;
  struct connection_info *c_info = p_data->c_info;

  switch ( pkt_header->packet_number )
    {
    case 0: /* "Initialization Packet" (greeting packet) */
      packet += 1 ; /* Skip protocol version */

      /* Keep server version */
      i = strlen((char *)packet) + 1;
      if ( ! (c_info->mysql_server_version = malloc(i)) )
	return;
      strcpy(c_info->mysql_server_version, (char *)packet);
      packet += i;

      packet += 4 + 8 + 1; /* Skip thread_id + scramble_buff + filler */

      /* Keep server_capabilities */
      memcpy(&(c_info->s_caps), packet, 2);
      packet += 2;
      break;
    case 1: /* "Client authentication packet" (login packet) */
      memcpy(&(c_info->c_caps), packet, 2); /* Keep client capabilities */
      packet += 2;

      if ( c_info->c_caps & MYSQL_CAPS_41 ) /* New protocol 4.1 */
	{
	  memcpy(&(c_info->c_ext_caps), packet, 2); /* Keep extended client capabilities */
	  packet += 2;

	  packet += 4 + 1 + 23; /* Skip max_packet_size + charset_number + filler */
	}
      else
	packet += 3; /* Skip max_packet_size */

      /* Keep username */
      i = strlen((char *)packet) + 1;
      if ( ! (c_info->mysql_username = malloc(i)) )
	return;
      strcpy(c_info->mysql_username, (char *)packet);
      packet += i;

      if ( c_info->c_caps & MYSQL_CAPS_41 && c_info->c_caps & MYSQL_CAPS_CWDB ) /* Protocol 4.1 and connect with db */
	{
	  /* Skip password */
	  i = strlen((char *)packet) + 1;
	  packet += i;

	  /* Keep database name */
	  i = strlen((char *)packet) + 1;
	  if ( ! (p_data->mysql_db = malloc(i)) )
	    return;
	  strcpy(p_data->mysql_db, (char *)packet);
	  packet += i;
	}
      break;
    case 2: /* "Server Response */
      p_data->mysql_response = (int)*packet;

      c_info->handshake = 1;

      if ( c_info->c_caps & MYSQL_CAPS_41 && c_info->c_caps & MYSQL_CAPS_CWDB ) /* Protocol 4.1 and connect with db */
	exportBucket(bkt, 0);
    }
}

static void
mysql_dissect_command(FlowHashBucket *bkt, struct packet_header *pkt_header,
		      struct plugin_data *p_data, u_char *packet)
{
  char **buff;
  int i, keep;

  keep = 0;
  switch ( pkt_header->packet_number )
    {
    case 0: /* Command Packet */
      p_data->mysql_command = *packet; /* Keep and skip command */
      packet++;

      switch ( p_data->mysql_command )
	{
	case MYSQL_COMM_INIT_DB: /* Use DB */
	  buff = &(p_data->mysql_db);
	  keep = 1;
	  break;
	case MYSQL_COMM_QUERY:
	  buff = &(p_data->mysql_query);
	  keep = 1;
	  break;
	}

      if ( keep )
	{
	  i = pkt_header->packet_length;
	  if ( ! (*buff = realloc(*buff, i)) )
	    return;
	  strncpy(*buff, (char *)packet, i-1);
	  (*buff)[i-1] = '\0';
	}
      break;
    case 1: /* Response Packet */
      p_data->mysql_response = (int)*packet;

      exportBucket(bkt, 0);
      p_data->mysql_command = MYSQL_NUM_COMM + 1; /* Unknown */
      if ( p_data->mysql_query != NULL )
	p_data->mysql_query = '\0';
      p_data->mysql_response = -1; /* Error */
      break;
    }
}
